import requests
import asyncio
import logging
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('MetricCollector')


class MetricCollector:
    _OFFLINE_PARTITIONS_MBEAN = {
        'name': 'OfflinePartitions',
        'mbean': 'kafka.controller:type=KafkaController,name=OfflinePartitionsCount'}
    _UNDER_REPLICATED_PARTITIONS_MBEAN = {
        'name': 'UnderReplicatedPartitions',
        'mbean': 'kafka.server:type=ReplicaManager,name=UnderReplicatedPartitions'}
    _PREFERRED_REPLICA_IMBALANCE_MBEAN = {
        'name': 'PreferredReplicaImbalance',
        'mbean': 'kafka.controller:name=PreferredReplicaImbalanceCount,type=KafkaController'}
    _JOLOKIA_PORT = 8778

    def __init__(self, zk: BukuExhibitor):
        self.zk = zk

    async def _get_metrics_from_broker(self, broker_id: int):
        broker_address = self.zk.get_broker_address(broker_id)
        data = {'broker_address': broker_address, 'broker_id': broker_id, 'metrics': {}}
        for metric in self.get_metric_mbeans():
            metric_fetched = False
            try:
                response = requests.get("http://{}:{}/jolokia/read/{}".format(
                    broker_address, self._JOLOKIA_PORT, metric['mbean']))
                if response.status_code == 200:
                    response_body = response.json()
                    if response_body.get('status') == 200:
                        if response_body.get('value', {}).get('Value') is not None:
                            data['metrics'][metric['name']] = response_body['value']['Value']
                            metric_fetched = True
                if not metric_fetched:
                    _LOG.error("Fetching metric {} for broker: {} failed. Response from broker: {}:{}".format(
                        metric['name'], broker_id, response.status_code, response.text))
            except Exception as e:
                _LOG.error("Fetching metric {} for broker {} failed".format(metric['name'], broker_id), exc_info=e)
        return data

    async def _get_metrics_from_brokers(self, broker_ids):
        metrics = []
        for broker_id in broker_ids:
            metrics.append(asyncio.ensure_future(self._get_metrics_from_broker(broker_id)))
        metrics = await asyncio.gather(*metrics)
        return metrics

    def get_metrics_from_brokers(self, broker_ids=None):
        """
        Get metrics for brokers in the cluster
        :param broker_ids: List of broker_ids to fetch metrics for
        :return: List of dictionaries containing metrics for each broker
        {
            "metrics": {...},
            "broker_id": int,
            "broker_address": str
        }
        """
        broker_ids = self.zk.get_broker_ids() if not broker_ids else broker_ids
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return loop.run_until_complete(self._get_metrics_from_brokers(broker_ids))
        except Exception as e:
            _LOG.error('Could not fetch metrics from brokers', exc_info=e)
        finally:
            loop.close()

    @classmethod
    def get_metric_mbeans(cls):
        return [
            cls._OFFLINE_PARTITIONS_MBEAN,
            cls._UNDER_REPLICATED_PARTITIONS_MBEAN,
            cls._PREFERRED_REPLICA_IMBALANCE_MBEAN
        ]



import copy
import logging

import yaml

from bubuku.aws import AWSResources
from bubuku.aws.cluster_config import ClusterConfig
from bubuku.aws.ip_address_allocator import IpAddressAllocator
from bubuku.aws.node import KAFKA_LOGS_EBS

_LOG = logging.getLogger('bubuku.aws.ec2_node')


class Ec2NodeLauncher(object):
    def __init__(self, aws: AWSResources, cluster_config: ClusterConfig):
        self.aws = aws
        self.cluster_config = cluster_config

    def _launch_instance(self, ip: str, subnet: dict, ami: object, security_group_id: str, iam_profile):
        #
        # Override any ephemeral volumes with NoDevice mapping,
        # otherwise auto-recovery alarm cannot be actually enabled.
        #
        _LOG.info('Overriding ephemeral volumes to be able to set up AWS auto recovery alarm ')
        block_devices = []
        for bd in ami.block_device_mappings:
            if 'Ebs' in bd:
                #
                # This has to be our root EBS.
                #
                # If the Encrypted flag is present, we have to delete
                # it even if it matches the actual snapshot setting,
                # otherwise amazon will complain rather loudly.
                #
                # Take a deep copy before deleting the key:
                #
                bd = copy.deepcopy(bd)

                root_ebs = bd['Ebs']
                if 'Encrypted' in root_ebs:
                    del (root_ebs['Encrypted'])

                block_devices.append(bd)
            else:
                # ignore any ephemeral volumes (aka. instance storage)
                block_devices.append({
                    'DeviceName': bd['DeviceName'],
                    'NoDevice': ''
                })

        user_data = self.cluster_config.get_user_data()
        user_data['volumes']['ebs']['/dev/xvdk'] = KAFKA_LOGS_EBS
        taupage_user_data = '#taupage-ami-config\n{}'.format(yaml.safe_dump(user_data))

        _LOG.info('Launching node %s in %s', ip, subnet['AvailabilityZone'])
        resp = self.aws.ec2_client.run_instances(
            ImageId=ami.id,
            MinCount=1,
            MaxCount=1,
            SecurityGroupIds=[security_group_id],
            UserData=taupage_user_data,
            InstanceType=self.cluster_config.get_instance_type(),
            SubnetId=subnet['SubnetId'],
            PrivateIpAddress=ip,
            BlockDeviceMappings=block_devices,
            IamInstanceProfile={'Arn': iam_profile['Arn']},
            DisableApiTermination=False,
            EbsOptimized=True)

        instance_id = resp['Instances'][0]['InstanceId']
        _LOG.info('Instance %s launched, waiting for it to initialize', instance_id)
        self.aws.ec2_client.create_tags(
            Resources=[instance_id],
            Tags=[
                {'Key': 'Name', 'Value': self.cluster_config.get_cluster_name()},
                {'Key': 'StackName', 'Value': self.cluster_config.get_cluster_name()}
            ]
        )

        return instance_id

    def create_auto_recovery_alarm(self, instance_id):
        _LOG.info('Creating AWS auto recovery alarm for %s', instance_id)
        alarm_actions = ['arn:aws:automate:{}:ec2:recover'.format(self.aws.region)]
        alarm_name = '{}-{}-auto-recover'.format(self.cluster_config.get_cluster_name(), instance_id)

        self.aws.cloudwatch_client.put_metric_alarm(
            AlarmName=alarm_name,
            AlarmActions=alarm_actions,
            MetricName='StatusCheckFailed_System',
            Namespace='AWS/EC2',
            Statistic='Minimum',
            Dimensions=[{
                'Name': 'InstanceId',
                'Value': instance_id
            }],
            Period=60,  # 1 minute
            EvaluationPeriods=2,
            Threshold=0,
            ComparisonOperator='GreaterThanThreshold')
        _LOG.info('Created alarm %s', alarm_name)

    def launch(self):
        _LOG.info('Preparing AWS configuration for ec2 instance creation')
        ip_address_allocator = IpAddressAllocator(self.aws, self.cluster_config)
        subnet, ip = ip_address_allocator.allocate_ip_addresses(1)[0]
        return self._launch_instance(ip,
                                     subnet,
                                     self._find_taupage_amis(),
                                     self._get_security_group_id(),
                                     self._get_instance_profile())

    def _get_instance_profile(self):
        profile_name = 'profile-{}'.format(self.cluster_config.get_cluster_name())
        profile = self.aws.iam_client.get_instance_profile(InstanceProfileName=profile_name)
        _LOG.info("IAM profile %s exists, using it", profile_name)
        return profile['InstanceProfile']

    def _get_security_group_id(self) -> dict:
        _LOG.info('Configuring security group ...')
        security_groups = self.aws.ec2_client.describe_security_groups(
            Filters=[{'Name': 'group-name', 'Values': [self.cluster_config.get_cluster_name()]}])
        if security_groups['SecurityGroups']:
            sg = security_groups['SecurityGroups'][0]
            _LOG.info('Security group for %s exists, will use it %s',
                      self.cluster_config.get_cluster_name(), sg['GroupId'])
            return sg['GroupId']
        raise Exception('Security group does not exist for {}'.format(self.cluster_config.get_cluster_name()))

    def _get_ip_permission(self, port: int):
        return {
            'IpProtocol': 'tcp',
            'FromPort': port,
            'ToPort': port,
            'IpRanges': [{'CidrIp': '0.0.0.0/0'}]
        }

    def _find_taupage_amis(self) -> dict:
        region = self.cluster_config.get_aws_region()
        _LOG.info('Finding latest Taupage AMI in %s..', region)

        filters = [{'Name': 'name', 'Values': ['*Taupage-AMI-*']},
                   {'Name': 'is-public', 'Values': ['false']},
                   {'Name': 'state', 'Values': ['available']},
                   {'Name': 'root-device-type', 'Values': ['ebs']}]
        images = list(self.aws.ec2_resource.images.filter(Filters=filters))
        if not images:
            raise Exception('No Taupage AMI found')
        most_recent_image = sorted(images, key=lambda i: i.name)[-1]

        _LOG.info('The latest AMI is %s', most_recent_image)

        return most_recent_image
